/*
*
*				Dombinations
*
*				Copyright 2007 	Harald Obermaier
*								http://www.zneaker.com
*
*/
module classes;

import SDL.SDL;
import std.stdio;
import std.random;
import dombinations;

//dimensions of window
const uint WIDTH 	= 400;	
const uint HEIGHT 	= 600;

const uint WHITE 	=	1;
const uint RED		=	2;


SDL_Surface *buffer = null;	//framebuffer
SDL_Surface *gui = null;	//Surface to hold data of the gui - bitmap

/* for non-static initialization of struct SDL_Rect */
void initRect(SDL_Rect *s,int x, int y, int w, int h)
{
	s.x = x;
	s.y = y;
	s.w = w;
	s.h = h;
}


//checks whether ix,iy is within the borders of rect
bool ON(int ix, int iy, SDL_Rect *rect) 
{
	return (ix>=rect.x && ix <= rect.x+rect.w && iy>= rect.y && iy <= rect.y+rect.h);
}




class Guess
{
private:
	int combination[4];	//Colors of this guess
	int rating[4];		//Colors of the rating of this guess

public:
	void reset()
	{
		for(int i = 0;i < 4;i++)
		{
			combination[i] = -1;
			rating[i] = 0;
		}
	}


	int rate(int c[4])
	{
		int red = 0, white = 0;	//amount of red / white "ratings"
		
		for(int i = 0;i < 4;i++)
			if(combination[i] == -1) //invalid combination -> do not rate it
				return -1;
				
		for(int i = 0;i < 4;i++)
			for(int u = 0;u < 4;u++)
				if(c[i] == combination[u])
				{
					if(i == u)	//this color is at the right position
						red++;
					else		//this color is part of the goal-combination - wrong position
						white++;
					break;
				}
				
		for(int i = 0;i < white;i++)
			rating[i] = WHITE;
		for(int i = white; i < white+red;i++)
			rating[i] = RED;
		for(int i = white+red;i < 4;i++)
			rating[i] = 0;
			
		if(red == 4) //guessed right combination
			return 1;
			
		return 0;
	}


	int getColor(int i)
	{
		return combination[i];	
	}


	void setColor(int i,int color)
	{
		combination[i] = color;	
	}


	int getRating(int i)
	{
		return rating[i];	
	}

}



class Dombinations
{

private:
	int mousex;		// mouse coordinates - x
	int mousey;		// mouse coordinates - y
	int win;		// 1 = victory, -1 = defeat, 0 = still playing
	uint starttime;	// time the game was started
	
	int combination[4];	// goal-combination
	int turn;			// number of current turn
	Guess guesses[10];	// 10 turns
	
public:
	this()
	{
		
		if(SDL_Init(SDL_INIT_VIDEO) == -1)
		{
			writef("Error: Could not initialize SDL.\n");
			done = true;
			return;
		}
		

		buffer = SDL_SetVideoMode(WIDTH,HEIGHT,16,SDL_SWSURFACE);//set video mode to WIDTHxHEIGHTx16
		if(!buffer)
		{
			writef("Error: Could not initialize videomode.\n");
			done = true;
			return;
		}
		//load bitmap
		gui = SDL_LoadBMP("gfx/gui.bmp");

		//get rid of that ugly background ( RGB(255,0,255) )
		SDL_SetColorKey(gui,SDL_SRCCOLORKEY,SDL_MapRGB(gui.format, 255, 0, 255));

		
		mousex = mousey = win =  0;
		
		for(int i = 0;i<10;i++)
			guesses[i] = new Guess();

		newGame();
	}

	~this()
	{
	}


	void newGame()
	{
		
		SDL_WM_SetCaption("Dombinations",null);
		for(int i = 0;i < 10;i++)
			guesses[i].reset();			//reset guesses
		
		for(int i = 0;i < 4;i++)
		{
			combination[i] = rand()%9;	//random color (0-8)
			for(int j = 0;j < i;j++)
			{
				if(combination[i] == combination[j])//no double-colors - generate color again
				{
					i--;
					break;	
				}
			}
		}
		
		win = turn = 0;
		starttime = SDL_GetTicks();	//save time of start
			
	}




	void render()
	{
		//white background
		static SDL_Rect screen = {0,0,WIDTH,HEIGHT};
		SDL_FillRect(buffer, &screen, SDL_MapRGB(buffer.format, 255, 255, 255));
		
		//render cells
		for(int i = 0;i<10;i++)
		{
			SDL_Rect cellb;
			initRect(&cellb,123,551-60*i-1,129,31);
			SDL_FillRect(buffer, &cellb, SDL_MapRGB(buffer.format, 150, 150, 150));
			for(int j = 0;j<4 ;j++)
			{
				SDL_Rect cellf;
				initRect(&cellf,124+32*j,551-60*i,31,29);
				if(ON(mousex,mousey,&cellf))
					SDL_FillRect(buffer, &cellf, SDL_MapRGB(buffer.format, 230, 230, 230));
				else
					SDL_FillRect(buffer, &cellf, SDL_MapRGB(buffer.format, 255, 255, 255));
			}
		}

		SDL_Rect srcrect;
		initRect(&srcrect,0,0,27,32);
		SDL_Rect dstrect;
		initRect(&dstrect,0,0,32,32);
		
		//render numbers
		for(int i = 0;i < 10;i++)
		{
			dstrect.y = 550 - 60*i;
			for(int j = 0; j < 4;j++)
			{
				dstrect.x = 125+32*j;
				if(guesses[i].getColor(j) != -1)
				{				
					srcrect.x = guesses[i].getColor(j)*27;
					SDL_BlitSurface(gui,&srcrect,buffer,&dstrect);
				}
											
			}
		}
		
		//render buttons

		if(!win)
		{
			{
				//randomize
				SDL_Rect dst1;
				initRect(&dst1,269,550-60*turn,30,30);	//border
				SDL_Rect dst2;
				initRect(&dst2,270,551-60*turn,28,28);
				SDL_FillRect(buffer, &dst1, SDL_MapRGB(buffer.format, 150, 150, 150));	//border
				if(ON(mousex,mousey,&dst2))
					SDL_FillRect(buffer, &dst2, SDL_MapRGB(buffer.format, 230, 230, 230));
				else
					SDL_FillRect(buffer, &dst2, SDL_MapRGB(buffer.format, 255, 255, 255));

				dstrect.x = 270;	
				dstrect.y = 550 - 60*turn;

				srcrect.x = 305;	
				SDL_BlitSurface(gui,&srcrect,buffer,&dstrect); //Randomize - Button
			
			}

			{
				//rate 
				SDL_Rect dst1;
				initRect(&dst1,309,550-60*turn,30,30);	//border
				SDL_Rect dst2;
				initRect(&dst2,310,551-60*turn,28,28);
				SDL_FillRect(buffer, &dst1, SDL_MapRGB(buffer.format, 150, 150, 150));	//border
				if(ON(mousex,mousey,&dst2))
					SDL_FillRect(buffer, &dst2, SDL_MapRGB(buffer.format, 230, 230, 230));
				else
					SDL_FillRect(buffer, &dst2, SDL_MapRGB(buffer.format, 255, 255, 255));

				dstrect.x = 314;	
				dstrect.y = 550 - 60*turn;

				srcrect.x = 333;	
				SDL_BlitSurface(gui,&srcrect,buffer,&dstrect);
			}
		}
		{
			//new game
			static SDL_Rect dst1 = {11,11,25,25};	//border
			static SDL_Rect dst2 = {12,12,23,23};
			SDL_FillRect(buffer, &dst1, SDL_MapRGB(buffer.format, 150, 150, 150));//border
			if(ON(mousex,mousey,&dst2))
				SDL_FillRect(buffer, &dst2, SDL_MapRGB(buffer.format, 230, 230, 230));
			else
				SDL_FillRect(buffer, &dst2, SDL_MapRGB(buffer.format, 255, 255, 255));

			dstrect.x = 10;	
			dstrect.y = 10;

			srcrect.x = 278;	
			SDL_BlitSurface(gui,&srcrect,buffer,&dstrect); 
		
		}
		


		
		
		if(win == -1)//we lost - display combination
		{
			dstrect.y = 10;
			for(int j = 0; j<4;j++)
			{
				dstrect.x = 267+32*j;		
				srcrect.x = combination[j]*27;
				SDL_BlitSurface(gui,&srcrect,buffer,&dstrect);		
			}
		}
		
		
		//display ratings
		srcrect.h = srcrect.w = dstrect.h = dstrect.w =10;
		for(int i = 0;i < 10;i++)
		{
			if(i>=turn)
				continue;
			for(int j = 0; j < 4;j++)
			{
				if(!guesses[i].getRating(j))
					break;
				dstrect.y = 600-50 - 60*i + 15*(j>1?1:0)+5;
				dstrect.x = 80+10*(j%2);
				
				srcrect.x = 239+guesses[i].getRating(j)*10;
				SDL_BlitSurface(gui,&srcrect,buffer,&dstrect); //ratings
											
			}
		}
		
		SDL_Flip(buffer);//swap buffers
		
		
	}


	void rate()
	{
		int rate = guesses[turn].rate(combination);
		if(!rate)
			turn++;	//combination is not the right one and we have still at least one turn left
			
		if(rate == 1)//combination is right
		{
			starttime = SDL_GetTicks()-starttime;
			//change title of window to: "Your time: [time]"
			char temp[64];
			sprintf(cast(char*)temp,"Your time:  %d : %d : %d",((starttime/1000)/60),(starttime/1000)%60,(starttime%1000)/10);
			SDL_WM_SetCaption(cast(char*)temp,null);
			win = 1;
			turn++;
		}
		else if(turn == 10) //we didn't get the right combination within 10 turns
			win = -1;
		
	}



	void randomize()
	{
		for(int i = 0;i < 4;i++)
		{
			int r = rand()%9;	//generate random number (0-8)
			int j = 0;
			for(;j < 4;j++)
			{
				if(guesses[turn].getColor(j) == r)//prevent combination from having a color twice
				{
					i--;
					break;
				}
			}
			if(j == 4)
				guesses[turn].setColor(i,r);
		}
	}

	void mouseDown(SDL_MouseButtonEvent click)
	{
		if(!win)
		{
			if(click.x >= 124 && click.x <= 249 && click.y >= 550-turn*60 && click.y <= 580-turn*60)
			{
				if(click.button == SDL_BUTTON_LEFT)	//increase number
					guesses[turn].setColor((click.x-124)/32,(guesses[turn].getColor((click.x-124)/32)+1)%9);
				else	//decrease number
				{
					guesses[turn].setColor((click.x-124)/32,((guesses[turn].getColor((click.x-124)/32) == -1) ? 8 : guesses[turn].getColor((click.x-124)/32)+8)%9);
				}
			}
			
		}
		
			
	}

	void mouseUp(SDL_MouseButtonEvent click)
	{
		if(!win)
		{
			{
				//randomize
				SDL_Rect dst1;
				initRect(&dst1,269,550-60*turn,30,30);
				if(ON(mousex,mousey,&dst1))
					randomize();
			
			}

			{
				//rate 
				SDL_Rect dst1;
				initRect(&dst1,309,550-60*turn,30,30);
				if(ON(mousex,mousey,&dst1))
					rate();
			}
		}
		{
			//new game
			static SDL_Rect dst1 = {11,11,25,25};
			if(ON(mousex,mousey,&dst1))
				newGame();
		
		}
	}

	void mouseMove(SDL_MouseMotionEvent move)
	{
		mousex = move.x;
		mousey = move.y;
	}

	int input()
	{
		SDL_Event event;
		while(SDL_PollEvent(&event))
		{
			switch(event.type)
			{
				case SDL_KEYDOWN:
					
					if(event.key.keysym.sym == SDLK_ESCAPE)
						return -1;
					else if(event.key.keysym.sym == SDLK_n)
						newGame();
					else if(!win)
					{
						if(event.key.keysym.sym == SDLK_r)
							randomize();
						else if(event.key.keysym.sym == SDLK_RETURN)
							rate();
					}
							
				break;
				case SDL_QUIT:
					return -1;
				break;
				case SDL_MOUSEBUTTONDOWN:
					mouseDown(event.button);
				break;
				case SDL_MOUSEBUTTONUP:
					mouseUp(event.button);
				break;
				case SDL_MOUSEMOTION:
					mouseMove(event.motion);
				break;
				default:
				break;
			}
		}
		return 0;	
	}
}
